/* Copyright (C) 2004 - 2007  db4objects Inc.  http://www.db4o.com

This file is part of the db4o open source object database.

db4o is free software; you can redistribute it and/or modify it under
the terms of version 2 of the GNU General Public License as published
by the Free Software Foundation and as clarified by db4objects' GPL 
interpretation policy, available at
http://www.db4o.com/about/company/legalpolicies/gplinterpretation/
Alternatively you can write to db4objects, Inc., 1900 S Norfolk Street,
Suite 350, San Mateo, CA 94403, USA.

db4o is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License along
with this program; if not, write to the Free Software Foundation, Inc.,
59 Temple Place - Suite 330, Boston, MA  02111-1307, USA. */
package com.db4o.internal;

import com.db4o.*;
import com.db4o.foundation.*;
import com.db4o.internal.handlers.*;
import com.db4o.internal.marshall.*;
import com.db4o.internal.query.processor.*;
import com.db4o.reflect.*;


/**
 * @exclude
 */
public class PrimitiveFieldHandler extends ClassMetadata{
    
    public final TypeHandler4 i_handler;
    
    PrimitiveFieldHandler(ObjectContainerBase a_stream, TypeHandler4 a_handler) {
    	super(a_stream, a_handler.classReflector());
        i_fields = FieldMetadata.EMPTY_ARRAY;
        i_handler = a_handler;
    }

    void activateFields(Transaction a_trans, Object a_object, int a_depth) {
        // Override
        // do nothing
    }


    final void addToIndex(LocalObjectContainer a_stream, Transaction a_trans, int a_id) {
        // Override
        // Primitive Indices will be created later.
    }

    boolean allowsQueries() {
        return false;
    }

    void cacheDirty(Collection4 col) {
        // do nothing
    }

    public boolean canHold(ReflectClass claxx) {
        // Do we need this at all???
        // Check if this method is ever called
        return i_handler.canHold(claxx);
    }
    
    public ReflectClass classReflector(){
    	return i_handler.classReflector();
    }
    
    public void deleteEmbedded(MarshallerFamily mf, StatefulBuffer a_bytes) {
        if(mf._primitive.useNormalClassRead()){
            super.deleteEmbedded(mf, a_bytes);
            return;
        }
        
        // Do nothing here, we should be in the payload area
        // no action to be taken.
    }

    
    public void deleteEmbedded1(MarshallerFamily mf, StatefulBuffer a_bytes, int a_id) {
        
        if(i_handler instanceof ArrayHandler){
            ArrayHandler ya = (ArrayHandler)i_handler;
            
            // TODO: the following checks, whether the array stores
            // primitives. There is one case that is not covered here:
            // If a primitive array is stored to an untyped array or
            // to an Object variable, they would need to be deleted 
            // and freed also. However, if they are untyped, every 
            // single one would have to be read an checked and this
            // would be extremely slow.
            
            // Solution: Store information, whether an object is 
            // primitive in our pointers, in the highest bit of the
            // length int.
            
            if(ya.i_isPrimitive){
                ya.deletePrimitiveEmbedded(a_bytes, this);
                a_bytes.slotDelete();
                return;
            }
        }
        
       if(i_handler instanceof UntypedFieldHandler){
            // Any-In-Any: Ignore delete 
            a_bytes.incrementOffset(i_handler.linkLength());
        }else{
            i_handler.deleteEmbedded(mf, a_bytes);
        }
		
		// TODO: Was this freeing call necessary? 
		//   free(a_bytes.getTransaction(), a_id, a_bytes.getAddress(), a_bytes.getLength());
		
		free(a_bytes, a_id);
			
    }

    void deleteMembers(MarshallerFamily mf, ObjectHeaderAttributes attributes, StatefulBuffer a_bytes, int a_type, boolean isUpdate) {
        if (a_type == Const4.TYPE_ARRAY) {
            new ArrayHandler(a_bytes.getStream(),this, true).deletePrimitiveEmbedded(a_bytes, this);
        } else if (a_type == Const4.TYPE_NARRAY) {
            new MultidimensionalArrayHandler(a_bytes.getStream(),this, true).deletePrimitiveEmbedded(a_bytes, this);
        }
    }
    
	final void free(Transaction a_trans, int a_id, int a_address, int a_length) {
        a_trans.slotFreePointerOnCommit(a_id, a_address, a_length);
	}
	
	final void free(StatefulBuffer a_bytes, int a_id) {
          a_bytes.getTransaction().slotFreePointerOnCommit(a_id, a_bytes.getAddress(), a_bytes.getLength());
	}
    
	public boolean hasIndex() {
	    return false;
	}

    Object instantiate(ObjectReference a_yapObject, Object a_object, MarshallerFamily mf, ObjectHeaderAttributes attributes, StatefulBuffer a_bytes, boolean a_addToIDTree) {
        if (a_object == null) {
            try {
                a_object = i_handler.read(mf, a_bytes, true);
            } catch (CorruptionException ce) {
                return null;
            }
            a_yapObject.setObjectWeak(a_bytes.getStream(), a_object);
        }
        a_yapObject.setStateClean();
        return a_object;
    }
    
    Object instantiateTransient(ObjectReference a_yapObject, Object a_object, MarshallerFamily mf, ObjectHeaderAttributes attributes, StatefulBuffer a_bytes) {
        try {
            return i_handler.read(mf, a_bytes, true);
        } catch (CorruptionException ce) {
            return null;
        }
    }

    void instantiateFields(ObjectReference a_yapObject, Object a_onObject, MarshallerFamily mf, ObjectHeaderAttributes attributes, StatefulBuffer a_bytes) {
        Object obj = null;
        try {
            obj = i_handler.read(mf, a_bytes, true);
        } catch (CorruptionException ce) {
            
        }
        if (obj != null) {
            i_handler.copyValue(obj, a_onObject);
        }
    }

    public boolean isArray() {
        return i_id == HandlerRegistry.ANY_ARRAY_ID || i_id == HandlerRegistry.ANY_ARRAY_N_ID;
    }
    
    public boolean isPrimitive(){
        return true;
    }
    
    public TernaryBool isSecondClass(){
        return TernaryBool.UNSPECIFIED;
    }
    
	public boolean isStrongTyped(){
		return false;
	}
    
    public void calculateLengths(Transaction trans, ObjectHeaderAttributes header, boolean topLevel, Object obj, boolean withIndirection) {
        i_handler.calculateLengths(trans, header, topLevel, obj, withIndirection);
    }
    
    public Comparable4 prepareComparison(Object a_constraint) {
        i_handler.prepareComparison(a_constraint);
        return i_handler;
    }
    
    public final ReflectClass primitiveClassReflector(){
        return i_handler.primitiveClassReflector();
    }
    
    public Object read(MarshallerFamily mf, StatefulBuffer a_bytes, boolean redirect) throws CorruptionException{
        if(mf._primitive.useNormalClassRead()){
            return super.read(mf, a_bytes, redirect);
        }
        return i_handler.read(mf, a_bytes, false);
    }

    public TypeHandler4 readArrayHandler(Transaction a_trans, MarshallerFamily mf, Buffer[] a_bytes) {
        if (isArray()) {
            return i_handler;
        }
        return null;
    }
    
    public Object readQuery(Transaction trans, MarshallerFamily mf, boolean withRedirection, Buffer reader, boolean toArray) throws CorruptionException{
        if(mf._primitive.useNormalClassRead()){
            return super.readQuery(trans, mf, withRedirection, reader, toArray);
        }
        return i_handler.readQuery(trans, mf, withRedirection, reader, toArray);
    }

    
    public QCandidate readSubCandidate(MarshallerFamily mf, Buffer reader, QCandidates candidates, boolean withIndirection) {
        return i_handler.readSubCandidate(mf, reader, candidates, withIndirection);
    } 

    void removeFromIndex(Transaction ta, int id) {
        // do nothing
    }
    
    public boolean supportsIndex() {
        return true;
    }

    public final boolean writeObjectBegin() {
        return false;
    }
    
    public Object writeNew(MarshallerFamily mf, Object a_object, boolean topLevel, StatefulBuffer a_bytes, boolean withIndirection, boolean restoreLinkOffset) {
        mf._primitive.writeNew(a_bytes.getTransaction(), this, a_object, topLevel, a_bytes, withIndirection, restoreLinkOffset);
        return a_object;
    }
    
    public String toString(){
        return "Wraps " + i_handler.toString() + " in YapClassPrimitive";
    }

    public void defrag(MarshallerFamily mf, ReaderPair readers, boolean redirect) {
        if(mf._primitive.useNormalClassRead()){
            super.defrag(mf,readers, redirect);
        }
        else {
            i_handler.defrag(mf, readers, false);
        }
    }
}
